// ****************************************************************************
// 
// 			Copyright  SCM Microsystems Inc.,2010
//			         All Rights Reserved
//
// ****************************************************************************
//
//     FileName : MCTEST.C
//
// ****************************************************************************
//
// Module Description
//
// A sample application to demonstrate the working of libmcscm library
//
// ****************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include "MCardAPI.h"
#include <assert.h>

#define VERSION_STRING "MCTEST for Linux Ver"
 
// List of supported memory card names
char gaszCardNames [][30] = \
{ 
  "UNKNOWN",
  "SLE 4406","SLE 4418","SLE 4428","SLE 4432","SLE 4436/5536/6636","SLE 4442","SLE 4436/5536/6636",
  "AT24C01ASC", "AT24C02SC",  "AT24C04SC",  "AT24C08SC",  "AT24C16SC", "AT24C32SC",
  "AT24C64SC",  "AT24C128SC", "AT24C256SC", "AT24C512SC", "AT88SC153",  "AT88SC1608" ,"SLE4404",
  "AT241024SC", "AT88SC1003", "ST1331", "ST14C02C","SLE 4436/5536/6636"
};

// File names used when writing to the memory Card 
char gaszWritefilenames [][30] = \
{
  "UNKNOWN", "sle4406.in", "sle4418.in", "sle4428.in", "sle4432.in", "sle4436.in", "sle4442.in",
  "sle5536.in", "at24c01asc.in", "at24c02sc.in", "at24c04sc.in", "at24c08sc.in", "at24c16sc.in",
  "at24c32sc.in", "at24c64sc.in", "at24c128sc.in", "at24c256sc.in",  "at24c512sc.in", "at88sc153.in",
  "at88sc1608.in", "sle4404.in", "at241024sc.in", "at88sc1003.in", "st1331.in", "st14c02c.in",
  "sle6636.in"
};

char gaszreadfilenames [][30] = \
{
  "UNKNOWN", "sle4406.out", "sle4418.out", "sle4428.out", "sle4432.out", "sle4436.out", "sle4442.out",
  "sle5536.out", "at24c01asc.out", "at24c02sc.out", "at24c04sc.out", "at24c08sc.out", "at24c16sc.out",
  "at24c32sc.out", "at24c64sc.out", "at24c128sc.out", "at24c256sc.out", "at24c512sc.out", "at88sc153.out",
  "at88sc1608.out", "sle4404.out", "at241024sc.out", "at88sc1003.out", "st1331.out", "st14c02c.out",
  "sle6636.out"	
 };

 // Protocol names
char gaszProtocolNames [][30] = {"UNKNOWN", "2 WIRE", "3 WIRE", "IIC", "EXT_IIC",  "AT88_IIC", "BIT"};

// Function for displaying menus
void ShowMenu  (void);
// Function for displaying card attributes
void ShowAttr  (void);
// Function that prints the error codes
void PutErrorCode  (LONG lStatus);
// Function to get the attributes
void GetAttribute  (MCARDHANDLE hMCard, int nAttrChoice);
// Function to handle any faults, interrupts
void SignalHandler  (int nVal);
// Function to validate inputs by user
BOOL CheckInput  (char *pcData);

int GetUserEntry  (void);


int main  (int argc, char **argv)
{
	SCARDCONTEXT hSCardContext = 0;
	MCARDHANDLE hMCard;
	MCARDCONTEXT hMCardContext;
	SCARD_READERSTATE_A rgReaderStates [1];
	
	BYTE abyInput [20] = {0};
	BYTE abyOutput [20] = {0};
	BYTE abyCurrentPIN [20] = {0};
	BYTE abyNewPIN [20] = {0};
	BYTE abyAttr [10] = {0};
 
	DWORD dwlength = 0, dwCurrentPINLength = 0, dwNewPINLength = 0;
	DWORD dwReaders = 0, dwVersion = 0, dwReadLen = 0;
	DWORD dwWriteLen = 0, dwLen = 0, dwOffset=0;
	DWORD dwDllVersion = 0, dwAttrLen = 0;

	BYTE byMemZone = 0, byPinNumber = 0;
	BYTE *pbyReadBuffer = NULL, *pbyWriteBuffer = NULL;
	BYTE *pbyMemStatus = NULL; 

	char szBuffer [50] = {0};
	char szInput[50] = {0};

	long lRetVal = 0;
	long lStatus = 0;
	
	BOOL bInit, bConnect, bVerify;
	
	char *pchReaders = NULL;
	char *pchTempPtr = NULL;
	const char *pchGroups = NULL;
	
	int anList [16] = {0};

	int nModeChoice = 0, nType = 0,	nindex = 0, nTemp = 0;
	int nMainReturn = 0, nReaderCount = 0, nCheckVal = 0;
	int nChoice = 1, nAttrChoice = 1, i = 0, p = 0, nReader = 0;

	do
	{
		// First, we capture all signals for proper exit
		signal  (SIGALRM, SignalHandler);
		signal  (SIGSEGV, SignalHandler);
		signal  (SIGILL, SignalHandler);
		signal  (SIGPIPE, SignalHandler);
	 	
		bInit = bConnect = bVerify = FALSE;

		dwVersion = APP_VERSION;

		printf  ("\n%s %d.%02d\n\n",VERSION_STRING, 
				 (int) ((dwVersion & 0x00FF0000) >> 16), 
				 (int) (dwVersion & 0x0000FFFF));

		printf ("Testing SCardEstablishContext    : ");
		
		// Establising the context
		lRetVal = SCardEstablishContext (SCARD_SCOPE_SYSTEM, NULL, NULL,&hSCardContext);

		printf ("%s\n", pcsc_stringify_error (lRetVal));

		if (lRetVal != SCARD_S_SUCCESS)
		{
			 nMainReturn = -1;
			 break;
		}

		printf ("Testing SCardGetStatusChange \n");
		printf ("Please insert a working reader   : ");
		
		// Checking for the reader presence
		lRetVal = SCardGetStatusChange  (hSCardContext, 0, 0, 0);

		printf ("%s\n", pcsc_stringify_error  (lRetVal));

		if (lRetVal != SCARD_S_SUCCESS)
		{
			SCardReleaseContext (hSCardContext);
			hSCardContext = 0;
			nMainReturn = -1;
			break;
		}

		printf ("Testing SCardListReaders         : ");

		pchGroups = 0;
		lRetVal = SCardListReaders (hSCardContext, pchGroups, 0, &dwReaders);

		printf ("%s\n", pcsc_stringify_error (lRetVal));

		if  (lRetVal != SCARD_S_SUCCESS)
		{
			SCardReleaseContext (hSCardContext);
			hSCardContext = 0;
			nMainReturn = -1;
			break;
		}

		pchReaders = (char *) malloc (sizeof (char) * dwReaders);

		lRetVal = SCardListReaders (hSCardContext, pchGroups, pchReaders, &dwReaders);

		if (lRetVal != SCARD_S_SUCCESS)
		{ 
			SCardReleaseContext (hSCardContext);
			hSCardContext = 0;
			nMainReturn = -1;
			break;
		}

		p = 0;

		for (i = 0; i < dwReaders - 1; i++)
		{
			++p;
			printf ("Reader %02d: %s\n", p, &pchReaders[i]);
			anList [p] = i;
			while (pchReaders [++i] != 0) ;
		}

		nReaderCount = p;

		do
		{
			printf ("\nEnter the desired reader number<1>  (-1 to exit): ");
	
			fgets (szBuffer,sizeof (szBuffer),stdin);

			//
			// here fgets reads the \n character also and after a \n it adds
			// '\0' in the string
			//

			if ('\n' == szBuffer[0])
			{
				nReader = 1;
				break;
			}

			//
			// Find the occurence of \n in the string
			//

			pchTempPtr = strstr (szBuffer,"\n");

			//
			// Replace that \n with a \0 to terminate the string there itself
			//

			*pchTempPtr = '\0';

			// The check here is added to validate for an 
			// exiting condition

			if  (0 == strcmp  ("-1", szBuffer))
			{
				nChoice = 15;
				break;
			}

			if (1 == strlen (szBuffer))
			{
				sscanf (szBuffer,"%d", &nReader);

				if  (nReader > nReaderCount || 0 >= nReader )
				{
					printf  ("Invalid Value - try again, -1 to exit\n");
				}
			}
			else
			{
				printf  ("Invalid Value - try again, -1 to exit\n");
			}
				 
				
		} while (nReader > nReaderCount || 0 >= nReader );
		
		// To support the above exitting condition where we set nChoice to 15

		if  (15 != nChoice)
		{
			rgReaderStates [0].szReader = &pchReaders [anList [nReader]];
			rgReaderStates [0].dwCurrentState = SCARD_STATE_EMPTY;
		}

		while (nChoice != 15)
		{
			ShowMenu ();
			 
			scanf  (" %[^\n]", szInput);
			
			if (strlen (szInput) > 2)
			{
				printf ("\nInvalid menu option\n");
				continue;
			}
			else
			{
				nChoice = -1;
				sscanf (szInput,"%d",&nChoice);

				switch (nChoice)
				{
					case 1:

						// Initializing the Mcard context

						lRetVal = MCardInitialize (hSCardContext, &pchReaders [anList [nReader]],
								&hMCardContext,	&dwDllVersion);

						if (SCARD_S_SUCCESS!=lRetVal)
						{
							 PutErrorCode (lRetVal);
						}
						else
						{
							bInit = TRUE;
							printf ("\nInitialization Done");
						}

						break;

					case 2:
						
						if (bInit)
						{
							// Getting the mcard handle by calling mcardconnect

							printf ("\n 1-To connect in Intelligent mode");
							printf ("\n 2-To connect in Forced mode");
							printf ("\n Enter your choice :");

							// This ignores any unused data in the buffer
							scanf (" %[^\n]", szInput);

							fflush (stdout);

			 				if (strlen (szInput) > 2)
							{
								printf ("\n Not a Valid Choice");
								continue;
							}

							sscanf (szInput,"%d", &nModeChoice);
								
							if (1 == nModeChoice)
							{					
								printf ("\nConnecting.....Please Wait!!!\n\n");

								lRetVal = MCardConnect (hMCardContext,
											INTELLIGENT_MODE, 0, &hMCard);
							}
							else if (2 == nModeChoice)
							{
								printf ("\nEnter the Type of Card as in the Specs :");
								scanf (" %x",&nType);
								printf ("\nConnecting.....Please Wait!!!\n\n");

								lRetVal = MCardConnect (hMCardContext, FORCED_MODE,
										nType, &hMCard);
							}
							else if (-1 == nModeChoice)
							{
								break;
							}
							else
							{
								printf ("\nNot a Valid choice");
								break;
							}


							if (MCARD_S_SUCCESS != lRetVal)
							{
								printf ("\nEither Card has been Removed or MCardInitialize not Called");
								break;
							}

							bConnect = TRUE;
							printf ("\nConnect Done");

						}
						else
						{
							printf  ("\nInitialization not done");
						}

						break;

					case 3:

						if (bInit && bConnect)
						{
							ShowAttr ();
							scanf (" %[^\n]", szInput);

			 				if (strlen (szInput) > 2)
							{
								printf ("\n Not a Valid Choice");
								continue;
							}
							 
							// Getting the attr code from the user
							sscanf (szInput, "%d", &nAttrChoice);

							if (-1 == nAttrChoice)
							{
								break;
							}
							
							if (nAttrChoice >= 1 && nAttrChoice <= 9)
							{
								GetAttribute (hMCard, nAttrChoice);
							}
							else
							{
								printf ("\nNot a valid choice");
							}
						}
						else
						{
							printf ("\n Either initialization or connect not done");
						}

						break;

					case 4:

						if (bInit && bConnect)
						{				
							// Getting the details to read from the user
							dwReadLen = 0;
							printf ("\nEnter the no of bytes to be Read (Hex)    :");

							scanf (" %[^\n]", szInput);

							if (CheckInput (szInput))
							{
								printf ("\nInvalid input data");
								break;
							}

							sscanf (szInput, "%x", &dwReadLen);

							// Invalid data like characters are handled here.

							if ((-1 == dwReadLen) || (0 == dwReadLen))
							{
								// This is there to flush the stdin 
								// from the \n that gets hanging around
								dwReadLen = GetUserEntry  ();
								break;
							}

							printf ("\nEnter the zone no           (Hex)         :");
							scanf (" %[^\n]", szInput);

							if (CheckInput (szInput))
							{
								printf ("\nInvalid input data");
								break;
							}

							sscanf (szInput, "%x", &nTemp);
							byMemZone = nTemp;
							printf ("\n Enter the byte Offset to Read  (Hex)     :");
							scanf (" %[^\n]", szInput);

							if (CheckInput (szInput))
							{
								printf ("\nInvalid input data");
								break;
							}

							sscanf (szInput, "%x", &nTemp);
							dwOffset = nTemp;
							pbyReadBuffer = (LPBYTE) malloc (dwReadLen*sizeof (BYTE));
							printf ("\n Reading Data from Card...Please wait\n\n");

							lRetVal = MCardReadMemory (hMCard,byMemZone,dwOffset, \
										pbyReadBuffer, &dwReadLen);

							if (MCARD_S_SUCCESS == lRetVal)
							{
								printf ("\nData Read successfully From the card\n");
								FILE *fpRead = NULL;
								dwAttrLen = 10;

								// Get the card type to obtain the index into the file names
								lRetVal = MCardGetAttrib (hMCard, MCARD_ATTR_TYPE, \
										abyAttr, &dwAttrLen);

								if (lRetVal != MCARD_S_SUCCESS)
								{
									printf ("Error in getting attribute type");
								}

								nindex = abyAttr[0];

								// Open the corrresponding file where the read bytes have to be stored

								fpRead = fopen (gaszreadfilenames [nindex], "wb");

								if (NULL == fpRead)
								{
									printf ("Error opening file for read dump\n");
								}

								printf (" Dumping to file [%s] \n", gaszreadfilenames[nindex]);

								for (i = 0; i < dwReadLen; i++)
								{
									fwrite (&pbyReadBuffer[i], sizeof (BYTE), 1, fpRead);
								}

								if (fpRead)
								{
									fclose (fpRead);
								}
							}
							else
							{
								PutErrorCode (lRetVal);
							}

							if (pbyReadBuffer) 
							{
								free (pbyReadBuffer);
								pbyReadBuffer = NULL; 
							}
						}
						else
						{
							printf ("\n Either initialization or Connect not done");
						}
							break;

					case 5:

						if (bInit && bConnect)
						{
							// Testing MCardWrite
							printf ("\nTesting MCard Write memory");
							printf ("\nEnter the no of bytes to be Written  (Hex)  :");

							scanf (" %[^\n]", szInput);

							if (CheckInput (szInput))
							{
								printf ("\nInvalid input data");
								break;
							}

							sscanf (szInput, "%x", &dwWriteLen);

							if (-1 == dwWriteLen)
							{
								break;
							}

							printf ("\nEnter the zone no       (Hex)              :");

							scanf (" %[^\n]", szInput);

							if (CheckInput (szInput))
							{
								printf ("\nInvalid input data");
								break;
							}

							sscanf (szInput, "%x", &nTemp);

							byMemZone = nTemp;
							printf ("\n Enter the byte Offset to write (Hex)      :");

							scanf (" %[^\n]", szInput);

							if (CheckInput (szInput))
							{
								printf ("\nInvalid input data");
								break;
							}

							sscanf  (szInput, "%x", &nTemp);
							dwOffset = nTemp;
							pbyWriteBuffer = (LPBYTE)malloc (dwWriteLen*sizeof (BYTE));
							assert (pbyWriteBuffer);
							FILE *fpWrite = NULL;

							// Get the type to obtain the index into the input file name
							lRetVal = MCardGetAttrib (hMCard, MCARD_ATTR_TYPE, abyAttr, &dwAttrLen);

							if (lRetVal != MCARD_S_SUCCESS)
							{
								printf (" Error in getting attribute type");
							}

							nindex = abyAttr[0];
							fpWrite = fopen (gaszWritefilenames [nindex], "rb");

							if (fpWrite == NULL)
							{
								printf ("\nFile: %s not found", gaszWritefilenames [nindex]);
								break;
							}

							BYTE byTempVar;
							DWORD dwCurrent = 0;

							while (1)
							{
								fread (&byTempVar,sizeof (BYTE),1,fpWrite);

								if (feof (fpWrite))
								{
									break;
								}
								else
								{
									if (dwCurrent<dwWriteLen)
									{
										pbyWriteBuffer[dwCurrent++] = byTempVar;
									}

								}
							}

							// We should only write what we have and not
							// more than that. so, make that check before
							// proceeding

							if  (dwCurrent < dwWriteLen)
							{
								printf  ("\n Not enough data to write");
								break;
							}

							printf  ("\n Writing to card ..Please Wait!\n\n");

							lRetVal = MCardWriteMemory (hMCard,byMemZone,dwOffset,pbyWriteBuffer,&dwWriteLen);

							if (lRetVal == MCARD_S_SUCCESS)
							{
								printf ("\n Written successfully");
							}
							else
							{
								PutErrorCode (lRetVal);
							}

							if (fpWrite)
							{
								fclose (fpWrite);
							}

							if (pbyWriteBuffer) 
							{
								free (pbyWriteBuffer);
								pbyWriteBuffer = NULL; 
							}
						}
						else
						{
							printf  ("\nEither Initialization or Connect not done");
						}

						break;

					case 6:

						if (bInit && bConnect)
						{
							int nPinLen = 0;
							DWORD dwAttrLen = 50;
							MCARD_FEATURES MCardFeatures;

							// PIN verification
		
							if  (MCARD_S_SUCCESS != MCardGetAttrib  (\
													hMCard,MCARD_ATTR_FEATURES,
													 (unsigned char *) &MCardFeatures,
													&dwAttrLen))
							{
								printf  ("\n Error in retrieving PIN Features.\n");
								break;
							}

							if  (!MCardFeatures.byPINs)
							{
								printf  ("\n No PIN features available.\n");
								break;
							}

							printf ("\n PIN Verification");

							printf ("\n Enter the pin number   : ");
							scanf  (" %[^\n]", szInput);

							if (CheckInput  (szInput))
							{
								printf  ("\nInvalid input data");
								break;
							}

							sscanf  (szInput, "%d", &byPinNumber);

							printf  (" Enter Pin length       : ");
							scanf  (" %[^\n]", szInput);

							if (CheckInput  (szInput))
							{
								printf ("\nInvalid input data");
								break;
							}

							sscanf  (szInput, "%d", &nPinLen);

							printf (" Enter the PIN          : ");

							fflush (stdin);
							scanf (" %[^\n]",abyInput);

							retHex (abyInput,abyOutput,&dwlength);

							if (nPinLen != dwlength)
							{
								printf  ("\nWrong Pin Length and data given\n");
								break;
							}

							lRetVal = MCardVerifyPIN (hMCard,byPinNumber,abyOutput,dwlength);

							if (lRetVal == MCARD_S_SUCCESS)
							{
								printf ("\nPIN verification:Success");
							}
							else
							{
								PutErrorCode (lRetVal);
								GetAttribute  (hMCard, 5);
							}
						}
						else
						{
							printf ("\n Either Initialization or Connect not done");
						}

						break;

					case 7:

						if (bInit && bConnect)
						{
							// For Non-PIN cards, we need not go and get
							// input data from users. quit before that

							DWORD dwAttrLen = 50;
							MCARD_FEATURES MCardFeatures;

							if  (MCARD_S_SUCCESS != MCardGetAttrib  (\
													hMCard,MCARD_ATTR_FEATURES,
													 (unsigned char *) &MCardFeatures,
													&dwAttrLen))
							{
								printf  ("\n Error in retrieving PIN Features.\n");
								break;
							}

							if  (!MCardFeatures.byPINs)
							{
								printf  ("\n No PIN features available.\n");
								break;
							}

							// Write code for PIN change
							printf ("\n PIN Change");
							printf ("\n Enter the pin number  :");
							scanf (" %[^\n]", szInput);

							if (CheckInput  (szInput))
							{
								printf  ("\nInvalid input data");
								break;
							}

							sscanf  (szInput, "%d", &byPinNumber);

							printf ("\n Enter the current PIN :");
							fflush (stdin);
							scanf (" %[^\n]",abyInput);

							retHex (abyInput,abyCurrentPIN,&dwCurrentPINLength);

							printf ("\n Enter the New PIN     :");

							fflush (stdin);
							scanf (" %[^\n]",abyInput);

							retHex (abyInput,abyNewPIN,&dwNewPINLength);

							if  (MCARD_S_SUCCESS !=  (lStatus = MCardChangePIN (hMCard,
											byPinNumber, abyCurrentPIN,	dwCurrentPINLength,
											abyNewPIN, dwNewPINLength)))
							{
								printf  ("\n Error in changing pin :");
								PutErrorCode  (lStatus);
							}
							else
							{
								printf ("\nThe PIN was successfully changed");
							}
						}
						else
						{
							printf ("\nEither Initialization or connect not done");
						}

						break;

					case 8:

						if (bInit && bConnect)
						{

							printf ("\nMemory protection status");
							printf ("\nEnter the no of bytes to be Checked (Hex)  :");
							scanf  (" %[^\n]", szInput);

							if (CheckInput  (szInput))
							{
								printf  ("\nInvalid input data");
								break;
							}

							sscanf  (szInput, "%x", &nTemp);

							if (-1 == nTemp)
							{
								break;
							}

							dwReadLen = nTemp;
							printf ("\nEnter the zone no  (Hex)                   :");
							scanf (" %[^\n]", szInput);

							if  (CheckInput (szInput))
							{
								printf  ("\nInvalid input data");
								break;
							}

							sscanf  (szInput, "%x", &nTemp);
							byMemZone = nTemp;
							printf ("\n Enter the byte Offset to Get Status (Hex) :");
							scanf (" %[^\n]", szInput);

							if (CheckInput  (szInput))
							{
								printf  ("\nInvalid input data");
								break;
							}

							sscanf  (szInput, "%x", &nTemp);
							dwOffset = nTemp;
							pbyMemStatus = (LPBYTE)malloc (dwReadLen*sizeof (BYTE));

							lRetVal = MCardReadMemoryStatus (hMCard,byMemZone,dwOffset,pbyMemStatus,&dwReadLen);

							if (MCARD_S_SUCCESS==lRetVal)
							{
								printf  ("\n Reading Status from Card....Please Wait!\n\n");

								for (i = 0; i < dwReadLen; i++)
								{
									printf ("%01x ", pbyMemStatus [i]);

									if (0 == (i + 1)%8)
									{
										printf ("\n");
									}
								}
							}
							else
							{
								printf  ("\n Error :");
								PutErrorCode (lRetVal);
							}

							if (pbyMemStatus) 
							{
								free (pbyMemStatus);
								pbyMemStatus = NULL; 
							}
						}
						else
						{
							printf ("\nEither Initialization or Connect not done");
						}

						break;

					case 9:

						printf ("\n Testing Challenge Response");
						BYTE	*pbyChallengeID;
						BYTE	*pbyChallengeLength;
						BYTE	byChallenge [10];
						BYTE	byResponse [10];
						BYTE	byResponseLength;

						int i;
						char szDispBuffer [100];
						unsigned long dwlength;

						// Allocating memory for challenge id and length
						// this is done dynamically coz on writing to contiguous locations as hex
						// the previous one gets over written

						pbyChallengeID =  (BYTE*)malloc (sizeof (BYTE));
						pbyChallengeLength =  (BYTE*)malloc (sizeof (BYTE));

						// Getting the challenge ID

						printf ("\n Enter the Challenge ID : ");
						scanf  (" %[^\n]", szInput);

						if (CheckInput  (szInput))
						{
							printf  ("\n Invalid input data");
							break;
						}

						if (strlen  (szInput) > sizeof (BYTE))
						{
							printf  ("\n Invalid input data");
							break;
						}

						sscanf  (szInput, "%x", pbyChallengeID);

						// Getting the Challenge Length
						printf ("\n Enter the Challenge Length : ");
						scanf  (" %[^\n]", szInput);

						if (CheckInput  (szInput))
						{
							printf  ("\nInvalid input data");
							break;
						}

						sscanf  (szInput, "%x", &dwlength);

						*pbyChallengeLength =  (BYTE) ( dwlength & 0xFF);

						// Getting the challenge from the corresponding control
						printf ("\n Enter the challenge  : ");
						fflush (stdin);
						scanf (" %[^\n]",abyInput);
						retHex (abyInput,byChallenge,&dwlength);

						byChallenge[dwlength] = '\0';

						if  (dwlength != *pbyChallengeLength)
						{
							printf  ("\n Wrong Challenge length/data pair\n");
							break;
						}

						printf (" Getting Response from the Card.....Please Wait!\n\n"); 

						if  (MCARD_S_SUCCESS !=  (lStatus = MCardChallengeResponse (hMCard,
								*pbyChallengeID, byChallenge, *pbyChallengeLength, byResponse,
								&byResponseLength)))
						{
							PutErrorCode  (lStatus);

						}
						else
						{
							char szTempBuffer [30];
							strcpy  (szDispBuffer, "");

							for  (i = 0; i < byResponseLength; i++)
							{
								sprintf (szTempBuffer,"%02x",byResponse[i]);
								strcat  (szDispBuffer, szTempBuffer);
								strcat  (szDispBuffer, "  ");
							}

							printf ("\n Response from the card  : %s",szDispBuffer);
						}

						if (pbyChallengeID) 
						{
							free (pbyChallengeID);
							pbyChallengeID = NULL; 
						}

						if (pbyChallengeLength) 
						{
							free (pbyChallengeLength); 
							pbyChallengeLength = NULL; 
						}

						break;

					case 10:

						printf ("\nSet attributes Test");
						BYTE abyAttr [100];
						int nSetChoice;
						printf ("\nSetting BIT order ");
						printf ("\n1-LSB");
						printf ("\n2-MSB");
						printf ("\n3-DEFAULT");
						printf ("\nEnter your choice :");

						scanf  (" %[^\n]", szInput);
						
						if (strlen  (szInput) > 2)
						{
							printf  ("\n Not a Valid Choice");
							continue;
						}
						 
						// Getting the attr code from the user
						sscanf  (szInput, "%d", &nSetChoice);

						if (1 == nSetChoice)
						{
							abyAttr [0] = LSB;
						}
						else if (2 == nSetChoice)
						{
							abyAttr [0] = MSB;
						}
						else if  (3 == nSetChoice)
						{
							abyAttr [0] = DEFAULT;
						}
						else
						{
							printf ("Invalid Bit order.");
							break;
						}

						if  (MCARD_S_SUCCESS != MCardSetAttrib (hMCard,	MCARD_ATTR_BIT_ORDER,
									abyAttr, 1))
						{
							printf ("\nConfig Info Set : Failed setting bit order.");	
							break;
						}

						printf  ("Enter the Clock rate  (Hex):");
						scanf  ("%x", &abyAttr[0]);

						if  (MCARD_S_SUCCESS != MCardSetAttrib (hMCard,	MCARD_ATTR_CLOCK,
									abyAttr, 1))
						{
							printf ("Config Info and bit order set : Failed in setting Clock rate.");	
							break;
						}

						break;

					case 11:

						printf ("\n Decrement Counter \n");
						DWORD	dwCounterVal;

						// Getting the units to be decremented

						printf ("\n Enter the units to be decremented in hex : ");

						// To handle/validate invalid data

						scanf (" %[^\n]", szInput);

						if (CheckInput  (szInput))
						{
							printf  ("\nInvalid input data");
							break;
						}

						sscanf  (szInput, "%x", &dwCounterVal);

						printf ("\n This operation will permanently decrement the value in the counter,");

						// Check point before proceeding
						// with this potentially DESTRUCTIVE procedure
						printf  ("\n Do you want to proceed  (y/n) : ");
						
						scanf (" %[^\n]", szInput);

						if (strlen  (szInput) > 1)
						{
							printf  ("\n Invalid input value\n");
							break;
						}

						sscanf (szInput, "%c", &nCheckVal);

						if (('n' == nCheckVal) ||  ('N' == nCheckVal))
						{
							break;
						}

						if (('y' != nCheckVal) &&  ('Y' != nCheckVal))
						{
							printf  ("\n Invalid input value\n");
							break;
						}

						if (MCARD_S_SUCCESS != (lStatus = MCardDeductCounter (
								hMCard,	0, dwCounterVal)))
						{
							PutErrorCode  (lStatus);
						}
						else
						{
							printf ("\nCounter successfully Decremented"); 
						}

						break;

					case 12:

						printf ("\n Set Memory write Protection");
						printf ("\nEnter the  MemoryLen  (Hex)       : ");
						scanf  (" %[^\n]", szInput);

						if (CheckInput  (szInput))
						{
							printf  ("\nInvalid input data");
							break;
						}

						sscanf (szInput, "%x", &nTemp);

						if (-1 == nTemp)
						{
							break;
						}

						dwLen = nTemp;
						printf ("\nEnter the zone no    (Hex)        : ");
						scanf (" %[^\n]", szInput);

						if (CheckInput (szInput))
						{
							printf  ("\nInvalid input data");
							break;
						}

						sscanf (szInput, "%x", &nTemp);
						byMemZone = nTemp;
						printf ("\n Enter the byte Offset (Hex)      :");
						scanf (" %[^\n]", szInput);

						if (CheckInput  (szInput))
						{
							printf  ("\nInvalid input data");
							break;
						}

						sscanf  (szInput, "%x", &nTemp);
						dwOffset = nTemp;

						// Check point before proceeding
						// with this potentially DESTRUCTIVE procedure

						printf ("\n This is a destructive operation. \n");
						printf  ("\n Do you want to proceed  (y/n) : ");
						
						scanf (" %[^\n]", szInput);

						if (strlen  (szInput) > 1)
						{
							printf  ("\n Invalid input value\n");
							break;
						}

						sscanf (szInput, "%c", &nCheckVal);

						if (('n' == nCheckVal) ||  ('N' == nCheckVal))
						{
							break;
						}

						if (('y' != nCheckVal) &&  ('Y' != nCheckVal))
						{
							printf  ("\n Invalid input value \n");
							break;
						}

						if (MCARD_S_SUCCESS != (lStatus = MCardSetMemoryWriteProtection  (hMCard,
										byMemZone, dwOffset, &dwLen)))
						{
							printf  ("\n Error :");
							PutErrorCode  (lStatus);
						}
						else
						{
							printf ("\nMemory area successfully write protected.");
						}

						break;	
						
					case 13:

						if (bConnect)
						{
							// Disconnecting from the card

							lRetVal = MCardDisconnect (hMCard,SCARD_RESET_CARD);

							if (lRetVal!=MCARD_S_SUCCESS)
							{
								printf ("\nError in disconnect");
							}
							else
							{
								printf ("\nDisconnected");
							}

							bConnect = FALSE;
						}
						else
						{
							printf ("\n Not yet connected");
						}

						break;

					// When shutdown is called directly, without
					// calling disconnect/shutdown menu options,
					// the reader is not switched back to smart card mode
					// hence, this makes sure we shutdown properly
					// and switch the reader back before we go out

					case 14:
					case 15:

						if  (bConnect)
						{
							MCardDisconnect  (hMCard, SCARD_RESET_CARD);
							printf  ("\n Disconnected");
							bConnect = FALSE;
						}

						if (bInit)
						{				
							// Shutting down
							lRetVal = MCardShutdown (hMCardContext);

							if (lRetVal != MCARD_S_SUCCESS)
							{
								printf  ("\n Error in MCard Shutdown");
							}
							else
							{
								printf  ("\n Shut down: Success");
							}

							bInit = FALSE;
						}
						else
						{
							if  (15 != nChoice)
							{
								printf  ("\n Not yet initialized");
							}
						}

						break;

					default:
						printf  ("\n Invalid menu option\n");
				}
			}
		}
	} while (FALSE); 

	if (hSCardContext != 0)
	{
		//Releasing the context
		SCardReleaseContext (hSCardContext);
	}
	
	printf ("\n");

	if (MCARD_S_SUCCESS == lRetVal)
	{
		printf ("MC Test Completed Successfully !\n");
	}

	if (pchReaders)
	{
		free (pchReaders);
		pchReaders = NULL;
	}

	return 0;
} // End of Main

// Added to handle any system-raised exceptions
// like faults or interrupts, to have a graceful exit.

void SignalHandler (int nVal)
{
	printf  (" Caught Error signal %d. Quitting\n\n", nVal);
	exit (1);
}

/*++
	Function to get the attribute for the current memory card
	Inputs:
		hMCard--Handle to the memory card obtained in connect
		nAttrChoice--the choice obtained from the user
++*/
	
void GetAttribute (MCARDHANDLE hMCard,int nAttrChoice)
{
	BYTE abyAttr [50] = {0};
	DWORD dwAttrLen = 10;
	MCARD_FEATURES MCardFeatures;
	MCARD_MEMORY MCardMemory [10] = {{0}};
	MCARD_PIN MCardPIN [20] = {{0}};
	MCARD_COUNTER MCardCounter[2] = {{0}};
	MCARD_CR MCardCR [5] = {{0}};

	char szDispBuffer [1024] = {0};
	long lRetVal;
	int i;

	switch (nAttrChoice)
	{
		case 1:
			
			// Getting the attributes for the card
			lRetVal = MCardGetAttrib (hMCard,MCARD_ATTR_TYPE,abyAttr,&dwAttrLen);

			if (lRetVal != MCARD_S_SUCCESS)
			{
				printf ("\n Error in getting attribute type\n");
				break;
			}
			 
			printf ("\n Type of card :%s",gaszCardNames [abyAttr[0]]);
			break;

		case 2:

			dwAttrLen = 50;

			lRetVal = MCardGetAttrib (hMCard,MCARD_ATTR_PROTOCOL,abyAttr,&dwAttrLen);

			if (lRetVal!=MCARD_S_SUCCESS)
			{
				printf ("\n Error in getting attributes protocol\n");
				break;
			} 

			printf ("Protocol:%s",gaszProtocolNames[abyAttr[0]]);
			break;

		case 3:
			
			dwAttrLen = 50;

			lRetVal = MCardGetAttrib (hMCard,MCARD_ATTR_FEATURES, (unsigned char *) &MCardFeatures,&dwAttrLen);

			if (lRetVal != MCARD_S_SUCCESS)
			{
				printf ("\n Error in getting attributes features\n");
				break;
			}
			
			printf ("\n_____Features______");
			printf ("\nZones   : %d",MCardFeatures.byMemoryZones);
			printf ("\nPINs    : %d",MCardFeatures.byPINs);
			printf ("\nCounters: %d",MCardFeatures.byCounters);
			printf ("\nCRs     : %d",MCardFeatures.byCRs);
			
			break;

		case 4:

			dwAttrLen = 50;

			if (MCARD_S_SUCCESS != MCardGetAttrib (hMCard,MCARD_ATTR_FEATURES, \
					(unsigned char *) &MCardFeatures,&dwAttrLen))
			{
				printf ("\n Error in retrieving Features.\n");
				break;
			}

			lRetVal = MCardGetAttrib (hMCard,MCARD_ATTR_MEMORY, (unsigned char *) &MCardMemory,&dwAttrLen);

			if (lRetVal!=MCARD_S_SUCCESS)
			{
				printf ("\n Error in getting attributes Memory\n");
				break;
			}

			printf ("\n Memory Attributes  (Hex)  :");

			char szTempBuffer [30];
			strcpy  (szDispBuffer, "");

			for  (i = 0; i < MCardFeatures.byMemoryZones; i++)
			{
				sprintf (szTempBuffer,"%x",MCardMemory[i].dwSize);
				strcat  (szDispBuffer, szTempBuffer);
				strcat  (szDispBuffer, "  ");
			}

			printf ("%s\n",szDispBuffer);
			break;

		case 5:
			dwAttrLen = 50;
			 
			// Getting the features first

			if  (MCARD_S_SUCCESS != MCardGetAttrib  (hMCard,MCARD_ATTR_FEATURES, \
					(unsigned char *) &MCardFeatures,&dwAttrLen))
			{
				printf ("\n Error in retrieving Features.\n");
				break;
			}
			
			if  (MCardFeatures.byPINs)
			{
				if  (MCARD_S_SUCCESS != MCardGetAttrib (hMCard,	MCARD_ATTR_PIN,
						 (unsigned char *) &MCardPIN, &dwAttrLen))
				{
					printf ("\n Error in retrieving PIN organization.\n");
					break;
				}

				char szTempBuffer [1024] = {0};
				int nTotalPin = 0;
				
				strcpy  (szDispBuffer, "");
				
				nTotalPin = MCardFeatures.byPINs;
				
				printf ("\n\n Total PINs : %d\n ", nTotalPin);

				for  (i = 0; i < MCardFeatures.byPINs; i++)
				{
					sprintf  (szTempBuffer, "\n Pin %02d: ", i);
					strcat  (szDispBuffer, szTempBuffer);

					strcat  (szDispBuffer, "   Length : ");
					sprintf (szTempBuffer,"%02X",MCardPIN[i].bySize);
					 
					strcat  (szDispBuffer, szTempBuffer);
					strcat  (szDispBuffer, "    Retries : ");
					
					if (1 == MCardPIN[i].byRetries)
					{
						sprintf  (szTempBuffer, "Only one retry left!!!");
					}
					else
					{
						sprintf (szTempBuffer,"%02X",MCardPIN[i].byRetries);
					}
					 
					strcat  (szDispBuffer, szTempBuffer);
					strcat  (szDispBuffer, "  ");

				}
				printf ("%s\n",szDispBuffer);
			}
			else
			{
				printf  ("\n No PIN attributes available\n");
			}

			break;
			
		case 6:

			dwAttrLen = 50;

			if  (MCARD_S_SUCCESS != MCardGetAttrib (hMCard,MCARD_ATTR_FEATURES, \
					(unsigned char *) &MCardFeatures,&dwAttrLen))
			{
				printf ("\n Error in retrieving Features.\n");
				break;
			}
			
			if  (MCardFeatures.byCounters)
			{
				if  (MCARD_S_SUCCESS != MCardGetAttrib (hMCard,	MCARD_ATTR_COUNTERS,
						 (unsigned char *) &MCardCounter, &dwAttrLen))
				{
					printf ("\n Error in retrieving memory organization.\n");
					break;
				}

				strcpy (szDispBuffer, "");

				for (i = 0; i < MCardFeatures.byCounters; i++)
				{
					sprintf (szTempBuffer,"%x",MCardCounter[i].dwSize);
					 
					strcat  (szDispBuffer, szTempBuffer);
					strcat  (szDispBuffer, "  ");

					sprintf (szTempBuffer,"%x",MCardCounter[i].dwUnits);
					 
					strcat  (szDispBuffer, szTempBuffer);
					strcat  (szDispBuffer, "  ");
				}

				printf ("\n Counter Details : %s\n",szDispBuffer);
			}
			else
			{
				printf  ("\n No counter data available\n");
			}

			break;

		case 7:
			dwAttrLen = 50;

			if  (MCARD_S_SUCCESS != MCardGetAttrib  (hMCard,MCARD_ATTR_FEATURES, \
					(unsigned char *) &MCardFeatures,&dwAttrLen))
			{
				printf ("\n Error in retrieving Features.\n");
				break;
			}
			
			if (MCardFeatures.byCRs)
			{
				if (MCARD_S_SUCCESS != MCardGetAttrib (hMCard, MCARD_ATTR_CR,
						 (unsigned char *) &MCardCR, &dwAttrLen))
				{
					printf ("\n Error in retrieving CR organization.\n");
					break;
				}

				char szTempBuffer[50];
				strcpy  (szDispBuffer, "");

				for  (i = 0; i < MCardFeatures.byCRs; i++)
				{
					sprintf (szTempBuffer,"%d",MCardCR[i].dwChallengeLen);
					 
					strcat  (szDispBuffer, szTempBuffer);
					strcat  (szDispBuffer, "  ");
					sprintf (szTempBuffer,"%d",MCardCR[i].dwResponseLen);
					 
					strcat  (szDispBuffer, szTempBuffer);
					strcat  (szDispBuffer, "  ");
					sprintf (szTempBuffer,"%d",MCardCR[i].byRetries);
					 
					strcat  (szDispBuffer, szTempBuffer);
					strcat  (szDispBuffer, "  ");
				}

				printf ("%s",szDispBuffer);
			}
			else
			{
				printf  ("\n No CR attribute data available\n");
			}

			break;
			
		case 8:
			dwAttrLen = 50;

			if (MCARD_S_SUCCESS != MCardGetAttrib  (hMCard,MCARD_ATTR_CLOCK,abyAttr,&dwAttrLen))
			{
				printf ("\n Error in retrieving clock rate.\n");
				break;
			}

			printf (" Clock Rate  (Hex) :%x", abyAttr[0]);

			break;

		case 9:

			dwAttrLen = 50;

			if  (MCARD_S_SUCCESS != MCardGetAttrib (hMCard,MCARD_ATTR_BIT_ORDER,abyAttr,&dwAttrLen))
			{
				printf ("\n Error in retrieving bit order.\n");
				break;
			}

			if (LSB == abyAttr[0])
			{
				printf ("Bit Order   :LSB");
			}
			else if (MSB == abyAttr[0])
			{
				printf ("Bit Order    :MSB");
			}
			else
			{
				printf ("Bit Order    :NOTSET");
			}

			break;

		case 10:

			dwAttrLen = 50;

			lRetVal = MCardGetAttrib (hMCard,MCARD_ATTR_CONFIGURATION,abyAttr,&dwAttrLen);

			if (lRetVal!=MCARD_S_SUCCESS)
			{
				printf ("\n Error in getting attributes Configuration\n");
				break;
			}

			if (0 == dwAttrLen)
			{
				printf  (" No Configuration Attributes data available\n");
				break;
			}

			printf ("\n Configuration Attributes       :");

			for (i = 0; i < dwAttrLen; i++)
			{
				printf ("%02x", abyAttr[i]);
			}
			printf  ("\n");
			break;
	}
	 
}

/*
 	Funtion to just display the mainmenu for the application
 */	

void ShowMenu (void)
{
	printf ("\n\t\t\t________MENU_______");
	printf ("\n    1-Initialize  ");
	printf ("\t\t\t    2-Connect & ID ");
	printf ("\n    3-Get Attributes");
	printf ("\t\t\t    4-Read Memory");
	printf ("\n    5-Write Memory");
	printf ("\t\t\t    6-Verify PIN");
	printf ("\n    7-Change PIN");
	printf ("\t\t\t    8-Get Memory Protection Status");
	printf ("\n    9-Challenge/Response");
	printf ("\t\t   10-Set Attributes");
	printf ("\n   11-Decrement Counter--DESTRUCTIVE");
	printf ("\t   12-Set Memory Write Protection--DESTRUCTIVE");
	printf ("\n   13-Disconnect   ");
	printf ("\t\t\t   14-Shutdown");
	printf ("\n   15-Exit    ");
	printf ("\n\n    Enter your choice  :");
}

/* 
	Funtion to display the attributes menu of the application
*/

void ShowAttr (void)
{
	printf ("\n ______ATTRIBUTES  MENU_____");
	printf ("\n     	1-Get Type   ");
	printf ("\n     	2-Get Protocol of the card");
	printf ("\n     	3-Get Features of the MCard");
	printf ("\n     	4-Get Attributes of Memory");
	printf ("\n     	5-Get Attributes of PIN");
	printf ("\n 	6-Get Attributes of Counter");
	printf ("\n	7-Get Challenge Response");
	printf ("\n 	8-Get Attributes of Clock");
	printf ("\n 	9-Get Bit order");

	printf ("\n\n	Enter your choice   :");
}

/*
	Description:
		Funtion that converts the string sent into array of Hex Bytes
	Inputs:
		str--string to be converted
		buf--returned bytes of hex
		plen--length of array of bytes returned
*/     	
	

void retHex (unsigned char *str,unsigned char *buf,unsigned long *plen)
{
	unsigned int i = 0,j = 0,k;
	unsigned char a;
	
	while (str[i]!='\n'&& str[i]!='\0')//till end of string
	{
		if (str[i]==32)
		{
			i++;
			continue;
		}
		if (str[i]>=48 && str[i]<=57)//if it is a no
		{
			buf[j] = str[i]-48;
		}
		if (str[i]==65 || str[i]==97)//if it is a,b,c,d,e,f storing corresponding values
		{
			buf[j] = 0xa;
		}
		if (str[i]==66 || str[i]==98)
		{
			buf[j] = 0xb;
		}
		if (str[i]==67 || str[i]==99)
		{
			buf[j] = 0xc;
		}
		if (str[i]==68 || str[i]==100)
		{
			buf[j] = 0xd;
		}
		if (str[i]==69 || str[i]==101)
		{
			buf[j] = 0xe;
		}
		if (str[i]==70 || str[i]==102)
		{
			buf[j] = 0xf;
		}

		i++;
		j++;
	}
	 
	for (k = 0; k < j; k = k + 1) // Combining the adjacent pair of elements in one
	{
		a = buf [k] << 4 | buf [k + 1];
		k++;
		buf [k / 2] = a;
		
	}
	j = j / 2;
	*plen = j;
}

// This function is used basically to retrieve the
// \n from the commandline which causes the menu 
// display to show twice

int GetUserEntry  (void)
{
	char    chTempStr [5] = {0};
	char    *pchTempPtr = NULL;
	int     nRetVal = -1;

	do
	{
		fgets (chTempStr, sizeof  (chTempStr), stdin);
		  
		if  ('\n' == chTempStr[0])
		{
			nRetVal = 99;
			break;
		}

		// Find the occurence of \n in the string
		// Replace it with '\0' to terminate the string

		pchTempPtr = strstr  (chTempStr,"\n");

		if  (NULL != pchTempPtr)
		{
			*pchTempPtr = '\0';
		}

		if  (0 == strcmp  (chTempStr, "x"))
		{
			nRetVal = 98;
			break;
		}

		sscanf  (chTempStr,"%d", &nRetVal);

	}while (FALSE);

	return nRetVal;
}

// Function for validing hex digit inputs from user

BOOL CheckInput  (char *pcData)
{
	int i;
	BOOL bIsXDigit = 0;

	for  (i = 0; i < strlen  (pcData); i++)
	{
		if  (0 == isxdigit  (pcData[i]))
		{
			bIsXDigit = 1;
			break;
		}
	} 

	return bIsXDigit;
}

/*++
	Function to Print the Error code
	Input:
		lStatus--Status code of Error
++*/	

void PutErrorCode  (LONG lStatus)
{
	switch  (lStatus)
	{
	case MCARD_S_SUCCESS:
		 printf ("\nMCARD_S_SUCCESS");
		break;

	case MCARD_E_INTERNAL_ERROR:
		printf ("\nAn Internal ERROR occured in the DLL\n");
		break;

	case MCARD_E_NOT_IMPLEMENTED:
		printf ("\nThis API has not been implemented in the DLL\n");
		break;

	case MCARD_E_NOT_INITIALIZED:
		 printf ("\nThe MCard Initialize was not called\n");
		break;

	case MCARD_E_UNKNOWN_CARD:
		 printf ("\nThe Memory Card could not be identified by the DLL\n");
		break;

	case MCARD_E_INVALID_PARAMETER:
		 printf ("\nAtleast one Parameter is invalid in this API call\n");
		break;

	case MCARD_E_CHAL_RESP_FAILED:
		 printf ("\nThe Challenge/Response failed\n");
		break;

	case MCARD_E_INVALID_MEMORY_RANGE:
		 printf ("\nThe Memory range specified is invalid\n");
		break;

	case MCARD_E_INVALID_MEMORY_ZONE_ID:
		 printf ("\nThe Memory Zone specified is invalid\n");
		break;

	case MCARD_E_INVALID_PIN_ID:
		 printf ("\nThe PIN ID is invalid\n");
		break;

	case MCARD_E_INVALID_CHAL_RESP_ID:
		 printf ("\nThe Challenge/Response ID is invalid\n");
		break;
		
	case MCARD_E_ERASURE_NEEDED:
		printf ("\nErasure to be done before this write\n");
		break;

	case MCARD_W_NOT_ALL_DATA_READ:
		 printf ("\nAll the requested data could not be read\n");
		break;

	case MCARD_W_NOT_ALL_DATA_WRITTEN:
		 printf ("\nNot all Data written to Card\n");
		break;

	case MCARD_W_PIN_VERIFY_NEEDED:
		 printf ("\nPIN Verification is required for this operation\n");
		break;

	case MCARD_W_PIN_VERIFY_FAILED:
		 printf ("\nPIN Verification FAILED!!!\n");
		break;

	case MCARD_W_NO_PIN_ATTEMPTS_LEFT:
		 
		 printf ("\nNo PIN retries are left for this card. Card might be blocked\n");
		break;

	case MCARD_W_NO_UNITS_TO_DECREMENT:
		 printf ("\nThere are no unts left in this card for decrementing\n");
		break;

	case MCARD_W_REMOVED_CARD:
		 printf ("\nThe Card has been removed\n");
		break;
		
	case MCARD_W_PROTECTED_AREA:
		printf ("\nCan't Write into Protected Area\n");
		break;
		
	case MCARD_W_BITORDER_CHANGED:
		printf ("\nRevert back to the original Bit Order and retry\n");
		break;
	
	case MCARD_E_PROTO_MISMATCH:
		printf  ("\nProtocol error while connecting to card\n");
		break;

	case MCARD_E_PROTOCOL_ERROR:
		printf  ("\nProtocol error during card access\n");
		break;

	case MCARD_E_INCOMPATIBLE_READER:
		printf  ("\nThis module does not work with the specified reader\n");
		break;

	case MCARD_E_BUFFER_TOO_SMALL:
		printf  ("\nThe buffer for return data is too small\n");
		break;

	case MCARD_W_TIMEOUT:
		printf  ("\nTimeout occurred\n");
		break;

	case MCARD_W_ABORTED:
		printf  ("\nCommand Aborted\n");
		break;
	
	default:
		 printf ("\nThis error code is not recognised by this app [%lX]\n", lStatus );
	}
}


